﻿#include "DescribeResponseChecker.h"
#include "LiveCapsule.h"
#include "RtspResponseInfo.h"
#include "DescribeResponseChecker.h"
#include "Strings.h"
#include <regex>

CDescribeResponseChecker::CDescribeResponseChecker(CLiveCapsule& facade): appFacade(facade) {}

bool lookLikeAbsoluteUrl(const std::string& uri)
{
  std::regex pattern("^[[:alpha:]]{2,}\\://");
  return std::regex_search(uri, pattern);
}

void updateIfaceTypeTheSame(const std::string& uri, int expectedType, const CLive555Wrapper& wrapper, bool& statusToUpdate)
{
  CUriInfo parsedUri(wrapper, uri.c_str());
  if (parsedUri.isParsed() && parsedUri.isAbsolute()) {
    statusToUpdate &= parsedUri.getHostAddressType() == expectedType;
  }
}

std::string stringifyIfaceType(int ifType)
{
  const char* stringifiedType = (ifType == AF_INET ? "IP4" : (ifType == AF_INET6 ? "IP6" : ""));
  return stringifiedType;
}

std::string stringifyAsIpType(int ifType)
{
  const char* stringifiedType = (ifType == AF_INET ? "IPv4" : (ifType == AF_INET6 ? "IPv6" : ""));
  return stringifiedType;
}

void  CDescribeResponseChecker::assertVerification(bool result, const std::string& message)
{
  if (!result) {
    appFacade.StepLog.AddLogText(message.c_str());
  }
}

void CDescribeResponseChecker::assertUriVerificationResult(bool result, int expectedType, const std::string& refName)
{
  assertVerification(result, std::string("The version of the IP address in the Uri returned by DUT (in property/SDP attrubute: \"") +
    refName + "\") is not expected. Expected version is " + stringifyAsIpType(expectedType));
}

void  CDescribeResponseChecker::assertIfTypeVerificationResult(bool result, int expectedType, const std::string& refName)
{
  assertVerification(result, std::string("The sub-field <addrtype> of SDP field \"") + refName + "\" has unexpected value. The expected value is \"" + stringifyIfaceType(expectedType) + "\"");
}

void  CDescribeResponseChecker::assertAddressVerificationResult(bool result, int expectedType, const std::string& refName)
{
  assertVerification(result, std::string("The version of the IP address returned by DUT (in SDP attrubute: \"") + refName + "\") is not expected. Expected version is " + stringifyAsIpType(expectedType));
}

bool CDescribeResponseChecker::checkUriIpAddressType(const std::string& uri, int expectedType, const std::string& refName)
{
  bool result = true;
  if (lookLikeAbsoluteUrl(uri)) {
    CUriInfo uriInfo(appFacade.Live555Wrapper, uri.c_str());
    if (uriInfo.isParsed()) {
      updateIfaceTypeTheSame(uri, expectedType, appFacade.Live555Wrapper, result);
      assertUriVerificationResult(result, expectedType, refName);
    }
    else {
      result = false;
      assertVerification(result, std::string("The Url returned by DUT (") +  uri + ") in property/SDP attrubute: \"" + refName + "\" can not be parsed");
    }
  }
  return result;
}

bool CDescribeResponseChecker::checkUriFromProperty(const std::string& propertyValue, const std::string& propertyName, const CUriInfo& masterUri)
{
  bool result = true;
  if (!propertyValue.empty()) {
    result = checkUriIpAddressType(propertyValue, masterUri.getHostAddressType(), propertyName);
  }
  return result;
}

std::string makeUriFromIp(const std::string& ip)
{
  std::string decoratedIp = ip;
  if (ip.find(":") != std::string::npos) {
    decoratedIp = std::string("[") + decoratedIp + "]";
  }
  return std::string("http://") + decoratedIp;
}

bool CDescribeResponseChecker::checkSdpInterfaceDescription(const std::vector<std::string>& preparsedPropValue, const std::string& attributeName, int ifType)
{
  bool result = true;
  auto expectedType = stringifyIfaceType(ifType);
  if (!expectedType.empty()) {
    auto ifMarker = std::find(preparsedPropValue.begin(), preparsedPropValue.end(), "IN");
    if (ifMarker != preparsedPropValue.end() && (ifMarker + 1) != preparsedPropValue.end()) {
      result = *(ifMarker + 1) == expectedType;
      assertIfTypeVerificationResult(result, ifType, attributeName);

      if (result && (ifMarker + 2) != preparsedPropValue.end()) {
        updateIfaceTypeTheSame(makeUriFromIp(*(ifMarker + 2)), ifType, appFacade.Live555Wrapper, result);
        assertAddressVerificationResult(result, ifType, attributeName);
      }
    }
  }
  return result;
}

bool CDescribeResponseChecker::checkSdpUri(const std::vector<std::string>& preparsedPropValue, const std::string& attributeName, int ifType)
{
  bool result = true;
  if (preparsedPropValue.size() > 0 && preparsedPropValue[0].size() > 5) {
    result = checkUriIpAddressType(preparsedPropValue[0], ifType, attributeName);
  }
  return result;
}

bool CDescribeResponseChecker::checkUriFromSdpProperty(const std::vector<std::string>& propertyValues,
  const std::string& attributeName,
  const CUriInfo& masterUri,
  bool (CDescribeResponseChecker::*checkProc)(const std::vector<std::string>& preparsedPropValue, const std::string& attributeName, int ifType))
{
  bool success = true;
  for (const auto& value : propertyValues) {
    auto parts = CStrings::split(value, " ");
    success &= (this->*checkProc)(parts, attributeName, masterUri.getHostAddressType());
    if (!success) {
      break;
    }
  }
  return success;
}

bool CDescribeResponseChecker::CheckSdpResult(const std::string& fullDescribeResponse, const char* sdpMessage)
{
  appFacade.StepLog.CreateNewStep("Check of IP address type in response to RTSP DESCRIBE");
  CUriInfo callUri(appFacade.Live555Wrapper, appFacade.FullConfig.ControlConnection.Uri.c_str());
  CRtspResponseInfo parsedResponse(fullDescribeResponse);

  bool success = checkUriFromProperty(parsedResponse.getRtspProperty("Content-Base"), "Content-Base", callUri) &&
    checkUriFromProperty(parsedResponse.getRtspProperty("Content-Location"), "Content-Location", callUri) &&
    checkUriFromSdpProperty(parsedResponse.getSdpProperties("c="), "c=", callUri, &CDescribeResponseChecker::checkSdpInterfaceDescription) &&
    checkUriFromSdpProperty(parsedResponse.getSdpProperties("o="), "o=", callUri, &CDescribeResponseChecker::checkSdpInterfaceDescription) &&
    checkUriFromSdpProperty(parsedResponse.getSdpProperties("a=control:"), "a=control:", callUri, &CDescribeResponseChecker::checkSdpUri);

  if (success) {
    appFacade.StepLog.SetStepResult();
  }
  else {
    appFacade.StepLog.SetStepResult(1, "Verification of IP address type is failed");
  }
  return success;
}
